@@ -6,24 +6,34 @@
 #include <array>
 #include <errno.h>
 #include <inttypes.h>
+#include <ios>
 #include <math.h>
 #include <optional>
 #include <png.h>
+#include <pngconf.h>
+#include <stdint.h>
+#include <stdio.h>
 #include <string.h>
+#include <string>
+#include <utility>
 #include <vector>
 
-#include "defaultinitvec.hpp"
+#include "diagnostics.hpp"
 #include "file.hpp"
 #include "helpers.hpp" // assume
+#include "verbosity.hpp"
 
+#include "gfx/flip.hpp"
 #include "gfx/main.hpp"
+#include "gfx/rgba.hpp"
+#include "gfx/warning.hpp"
 
-static DefaultInitVec<uint8_t> readInto(std::string const &path) {
+static std::vector<uint8_t> readInto(std::string const &path) {
 	File file;
 	if (!file.open(path, std::ios::in | std::ios::binary)) {
 		fatal("Failed to open \"%s\": %s", file.c_str(path), strerror(errno));
 	}
-	DefaultInitVec<uint8_t> data(128 * 16); // Begin with some room pre-allocated
+	std::vector<uint8_t> data(128 * 16); // Begin with some room pre-allocated
 
 	size_t curSize = 0;
 	for (;;) {
@@ -52,27 +62,27 @@ static DefaultInitVec<uint8_t> readInto(std::string const &path) {
 [[noreturn]]
 static void pngError(png_structp png, char const *msg) {
 	fatal(
-	    "Error writing reversed image (\"%s\"): %s",
-	    static_cast<char const *>(png_get_error_ptr(png)),
+	    "libpng error while writing reversed image (\"%s\"): %s",
+	    reinterpret_cast<char const *>(png_get_error_ptr(png)),
 	    msg
 	);
 }
 
 static void pngWarning(png_structp png, char const *msg) {
-	warning(
-	    "While writing reversed image (\"%s\"): %s",
-	    static_cast<char const *>(png_get_error_ptr(png)),
+	warnx(
+	    "libpng found while writing reversed image (\"%s\"): %s",
+	    reinterpret_cast<char const *>(png_get_error_ptr(png)),
 	    msg
 	);
 }
 
 static void writePng(png_structp png, png_bytep data, size_t length) {
-	auto &pngFile = *static_cast<File *>(png_get_io_ptr(png));
+	File &pngFile = *static_cast<File *>(png_get_io_ptr(png));
 	pngFile->sputn(reinterpret_cast<char *>(data), length);
 }
 
 static void flushPng(png_structp png) {
-	auto &pngFile = *static_cast<File *>(png_get_io_ptr(png));
+	File &pngFile = *static_cast<File *>(png_get_io_ptr(png));
 	pngFile->pubsync();
 }
 
@@ -97,37 +107,37 @@ static void printPalette(std::array<std::optional<Rgba>, 4> const &palette) {
 }
 
 void reverse() {
-	options.verbosePrint(Options::VERB_CFG, "Using libpng %s\n", png_get_libpng_ver(nullptr));
+	verbosePrint(VERB_CONFIG, "Using libpng %s\n", png_get_libpng_ver(nullptr));
 
 	// Check for weird flag combinations
 
 	if (options.output.empty()) {
-		fatal("Tile data must be provided when reversing an image!");
+		fatal("Tile data must be provided when reversing an image");
 	}
 
 	if (options.allowDedup && options.tilemap.empty()) {
-		warning("Tile deduplication is enabled, but no tilemap is provided?");
+		warnx("Tile deduplication is enabled, but no tilemap is provided");
 	}
 
 	if (options.useColorCurve) {
-		warning("The color curve is not yet supported in reverse mode...");
+		warnx("The color curve is not yet supported in reverse mode");
 	}
 
 	if (options.inputSlice.left != 0 || options.inputSlice.top != 0
 	    || options.inputSlice.height != 0) {
-		warning("\"Sliced-off\" pixels are ignored in reverse mode");
+		warnx("\"Sliced-off\" pixels are ignored in reverse mode");
 	}
 	if (options.inputSlice.width != 0 && options.inputSlice.width != options.reversedWidth * 8) {
-		warning(
+		warnx(
 		    "Specified input slice width (%" PRIu16
-		    ") doesn't match provided reversing width (%" PRIu16 " * 8)",
+		    ") does not match provided reversing width (%" PRIu16 " * 8)",
 		    options.inputSlice.width,
 		    options.reversedWidth
 		);
 	}
 
-	options.verbosePrint(Options::VERB_LOG_ACT, "Reading tiles...\n");
-	auto const tiles = readInto(options.output);
+	verbosePrint(VERB_NOTICE, "Reading tiles...\n");
+	std::vector<uint8_t> const tiles = readInto(options.output);
 	uint8_t tileSize = 8 * options.bitDepth;
 	if (tiles.size() % tileSize != 0) {
 		fatal(
@@ -139,20 +149,20 @@ void reverse() {
 
 	// By default, assume tiles are not deduplicated, and add the (allegedly) trimmed tiles
 	size_t const nbTiles = tiles.size() / tileSize;
-	options.verbosePrint(Options::VERB_INTERM, "Read %zu tiles.\n", nbTiles);
+	verbosePrint(VERB_INFO, "Read %zu tiles\n", nbTiles);
 	size_t mapSize = nbTiles + options.trim; // Image size in tiles
-	std::optional<DefaultInitVec<uint8_t>> tilemap;
+	std::optional<std::vector<uint8_t>> tilemap;
 	if (!options.tilemap.empty()) {
 		tilemap = readInto(options.tilemap);
 		mapSize = tilemap->size();
-		options.verbosePrint(Options::VERB_INTERM, "Read %zu tilemap entries.\n", mapSize);
+		verbosePrint(VERB_INFO, "Read %zu tilemap entries\n", mapSize);
 	}
 
 	if (mapSize == 0) {
 		fatal("Cannot generate empty image");
 	}
 	if (mapSize > options.maxNbTiles[0] + options.maxNbTiles[1]) {
-		warning(
+		warnx(
 		    "Total number of tiles (%zu) is more than the limit of %" PRIu16 " + %" PRIu16,
 		    mapSize,
 		    options.maxNbTiles[0],
@@ -171,13 +181,13 @@ void reverse() {
 				break;
 			}
 		}
-		options.verbosePrint(Options::VERB_INTERM, "Picked reversing width of %zu tiles\n", width);
+		verbosePrint(VERB_INFO, "Picked reversing width of %zu tiles\n", width);
 	}
 	if (mapSize % width != 0) {
 		if (options.trim == 0 && !tilemap) {
 			fatal(
 			    "Total number of tiles (%zu) cannot be divided by image width (%zu tiles)\n"
-			    "(To proceed anyway with this image width, try passing `-x %zu`)",
+			    "(To proceed anyway with this image width, try passing \"-x %zu\")",
 			    mapSize,
 			    width,
 			    width - mapSize % width
@@ -191,9 +201,7 @@ void reverse() {
 	}
 	height = mapSize / width;
 
-	options.verbosePrint(
-	    Options::VERB_INTERM, "Reversed image dimensions: %zux%zu tiles\n", width, height
-	);
+	verbosePrint(VERB_INFO, "Reversed image dimensions: %zux%zu tiles\n", width, height);
 
 	Rgba const grayColors[4] = {
 	    Rgba(0xFFFFFFFF), Rgba(0xAAAAAAFF), Rgba(0x555555FF), Rgba(0x000000FF)
@@ -210,31 +218,31 @@ void reverse() {
 
 		palettes.clear();
 		std::array<uint8_t, sizeof(uint16_t) * 4> buf; // 4 colors
-		size_t nbRead;
-		do {
-			nbRead = file->sgetn(reinterpret_cast<char *>(buf.data()), buf.size());
-			if (nbRead == buf.size()) {
-				// Expand the colors
-				auto &palette = palettes.emplace_back();
-				std::generate(
-				    palette.begin(),
-				    palette.begin() + options.nbColorsPerPal,
-				    [&buf, i = 0]() mutable {
-					    i += 2;
-					    return Rgba::fromCGBColor(buf[i - 2] + (buf[i - 1] << 8));
-				    }
-				);
-			} else if (nbRead != 0) {
+		for (;;) {
+			if (size_t nbRead = file->sgetn(reinterpret_cast<char *>(buf.data()), buf.size());
+			    nbRead == 0) {
+				break;
+			} else if (nbRead != buf.size()) {
 				fatal(
-				    "Palette data size (%zu) is not a multiple of %zu bytes!\n",
+				    "Palette data size (%zu) is not a multiple of %zu bytes\n",
 				    palettes.size() * buf.size() + nbRead,
 				    buf.size()
 				);
 			}
-		} while (nbRead != 0);
+			// Expand the colors
+			auto &palette = palettes.emplace_back();
+			std::generate(
+			    palette.begin(),
+			    palette.begin() + options.nbColorsPerPal,
+			    [&buf, i = 0]() mutable {
+				    i += 2;
+				    return Rgba::fromCGBColor(buf[i - 2] | buf[i - 1] << 8); // little-endian
+			    }
+			);
+		}
 
 		if (palettes.size() > options.nbPalettes) {
-			warning(
+			warnx(
 			    "Read %zu palettes, more than the specified limit of %" PRIu16,
 			    palettes.size(),
 			    options.nbPalettes
@@ -242,9 +250,9 @@ void reverse() {
 		}
 
 		if (options.palSpecType == Options::EXPLICIT && palettes != options.palSpec) {
-			warning("Colors in the palette file do not match those specified with `-c`!");
+			warnx("Colors in the palette file do not match those specified with '-c'");
 			// This spacing aligns "...versus with `-c`" above the column of `-c` palettes
-			fputs("Colors specified in the palette file:         ...versus with `-c`:\n", stderr);
+			fputs("Colors specified in the palette file:         ...versus with '-c':\n", stderr);
 			for (size_t i = 0; i < palettes.size() && i < options.palSpec.size(); ++i) {
 				if (i < palettes.size()) {
 					printPalette(palettes[i]);
@@ -263,19 +271,19 @@ void reverse() {
 			palettes[0][i] = grayColors[options.dmgValue(i)];
 		}
 	} else if (options.palSpecType == Options::EMBEDDED) {
-		warning("An embedded palette was requested, but no palette file was specified; ignoring "
-		        "request.");
+		warnx("An embedded palette was requested, but no palette file was specified; ignoring "
+		      "request");
 	} else if (options.palSpecType == Options::EXPLICIT) {
 		palettes = std::move(options.palSpec); // We won't be using it again.
 	}
 
-	std::optional<DefaultInitVec<uint8_t>> attrmap;
+	std::optional<std::vector<uint8_t>> attrmap;
 	uint16_t nbTilesInBank[2] = {0, 0}; // Only used if there is an attrmap.
 	if (!options.attrmap.empty()) {
 		attrmap = readInto(options.attrmap);
 		if (attrmap->size() != mapSize) {
 			fatal(
-			    "Attribute map size (%zu tiles) doesn't match image's (%zu)",
+			    "Attribute map size (%zu tiles) does not match image size (%zu tiles)",
 			    attrmap->size(),
 			    mapSize
 			);
@@ -289,10 +297,11 @@ void reverse() {
 			uint8_t attr = (*attrmap)[index];
 			size_t tx = index % width, ty = index / width;
 
-			if ((attr & 0b111) > palettes.size()) {
+			if (uint8_t palID = (attr & 0b111) - options.basePalID; palID > palettes.size()) {
 				error(
-				    "Attribute map references palette #%u at (%zu, %zu), but there are only %zu!",
-				    attr & 0b111,
+				    "Attribute map references palette #%u at (%zu, %zu), but there are only %zu "
+				    "palettes",
+				    palID,
 				    tx,
 				    ty,
 				    palettes.size()
@@ -303,7 +312,7 @@ void reverse() {
 
 			if (!tilemap) {
 				if (bank) {
-					warning(
+					warnx(
 					    "Attribute map assigns tile at (%zu, %zu) to bank 1, but no tilemap "
 					    "specified; "
 					    "ignoring the bank bit",
@@ -319,8 +328,8 @@ void reverse() {
 			}
 		}
 
-		options.verbosePrint(
-		    Options::VERB_INTERM,
+		verbosePrint(
+		    VERB_INFO,
 		    "Number of tiles in bank {0: %" PRIu16 ", 1: %" PRIu16 "}\n",
 		    nbTilesInBank[0],
 		    nbTilesInBank[1]
@@ -356,7 +365,7 @@ void reverse() {
 				size_t tx = index % width, ty = index / width;
 				uint8_t tileID = (*tilemap)[index];
 				uint8_t attr = (*attrmap)[index];
-				bool bank = attr & 0x08;
+				bool bank = attr & 0b1000;
 
 				if (uint8_t tileOfs = tileID - options.baseTileIDs[bank];
 				    tileOfs >= options.maxNbTiles[bank]) {
@@ -391,19 +400,19 @@ void reverse() {
 		requireZeroErrors();
 	}
 
-	std::optional<DefaultInitVec<uint8_t>> palmap;
+	std::optional<std::vector<uint8_t>> palmap;
 	if (!options.palmap.empty()) {
 		palmap = readInto(options.palmap);
 		if (palmap->size() != mapSize) {
 			fatal(
-			    "Palette map size (%zu tiles) doesn't match image size (%zu)",
+			    "Palette map size (%zu tiles) does not match image size (%zu tiles)",
 			    palmap->size(),
 			    mapSize
 			);
 		}
 	}
 
-	options.verbosePrint(Options::VERB_LOG_ACT, "Writing image...\n");
+	verbosePrint(VERB_NOTICE, "Writing image...\n");
 	File pngFile;
 	if (!pngFile.open(options.input, std::ios::out | std::ios::binary)) {
 		// LCOV_EXCL_START
@@ -511,7 +520,7 @@ void reverse() {
 			            : index;
 			// This should have been enforced by the earlier checking.
 			assume(tileOfs < nbTiles + options.trim);
-			size_t palID = palmap ? (*palmap)[index] : attribute & 0b111;
+			size_t palID = (palmap ? (*palmap)[index] : attribute & 0b111) - options.basePalID;
 			assume(palID < palettes.size()); // Should be ensured on data read
 
 			// We do not have data for tiles trimmed with `-x`, so assume they are "blank"
